import java.awt.*; //Label, Image, Graphics, Color
import java.applet.*; //Applet
import java.io.*; //DataInputStream
import java.net.*; //URL
import java.awt.event.*; //KeyListener

public class Java3DEngines extends Applet implements Runnable, KeyListener
{
	//Window objects
	Image workspace;
	Graphics offscreen;

	//Application options
	int MovieWidth, MovieHeight; //Movie dimensions
	Color PolyColor, WireColor; //Draw colors
	boolean WireFrame, Polys, BackFaceCulling; //Toggle options

	//Mesh information
	int FaceLen=0; //If FaceLen stays 0, we know data was never loaded
	float VertexArray[];
	int FaceArray[];
	
	//3D calculation information
	float matWorld[][] = new float[][] {{1,0,0,0}, {0,1,0,0}, {0,0,1,0}, {0,0,0,1}};
	float matViewProj[][], matFinal[][], CameraNorm[];
	float CurRotate=0;
	
	//Turn a movie parameter into an integer
	public int GP(String ParamName) { return GP(ParamName, 10); }
	public int GP(String ParamName, int Radix) { return Integer.parseInt(getParameter(ParamName), Radix); }

	//Initialization all data
	public void init()
	{
		//Get passed movie parameters
		MovieWidth = GP("Width");
		MovieHeight = GP("Height");
		this.setBackground(new Color(GP("background", 16)));
		PolyColor=new Color(GP("PolyColor", 16));
		WireColor=new Color(GP("WireFrameColor", 16));
		WireFrame=GP("ShowWireFrame")!=0;
		Polys=GP("ShowPolys")!=0;
		BackFaceCulling=GP("BackFaceCulling")!=0;
		
		//Create window layout
		addKeyListener(this); //Listen for key strokes
		workspace = createImage(MovieWidth, MovieHeight);
		offscreen = workspace.getGraphics();
		offscreen.drawString("Loading...", 10, 20);

		//Calculate combined world, view, and projection data
		float matView[][] = new float[][] {{1,0,0,0}, {0,0.857493f,-0.514496f,0}, {0,0.514496f,0.857493f,0}, {0,0,5.83095f,1}};
		float matProj[][] = new float[][] {{2.41421f,0,0,0}, {0,2.41421f,0,0}, {0,0,1.00100f,1}, {0,0,-0.0100100f,0}};
		matViewProj=MatrixMatrixMultiply(matView,matProj);
		CameraNorm=Vec3Sub(new float[] {0.0f, 3.0f,-5.0f},new float[] {0.0f, 0.0f, 0.0f});

		try
		{
			//Attempt to open model data file
			URLConnection conn=new URL(getParameter("InputFile")).openConnection();
			conn.connect();
			BufferedInputStream BufStream = new BufferedInputStream(conn.getInputStream());			
			DataInputStream data = new DataInputStream(BufStream);
			
			//Read in number of verticies and vertex information
			int NumVertsLcl=ReadInLittleEndian(data, 2);
			VertexArray=new float[NumVertsLcl*3];
			for(int i=0;i<NumVertsLcl*3;i++)
				VertexArray[i]=Float.intBitsToFloat(ReadInLittleEndian(data, 4));
			
			//Read in number of faces and face (index) information
			FaceLen=ReadInLittleEndian(data, 2);
			FaceArray=new int[FaceLen*3];
			for(int i=0;i<FaceLen*3;i++)
				FaceArray[i]=(int)ReadInLittleEndian(data, 2);
		}
		catch(IOException e1)
		{
			offscreen.drawString("IO Exception: "+e1.toString(), 10, 20);
		}
	}
	
	//Read in unsigned little endian integer because DataInputString only reads in big endian
	public int ReadInLittleEndian(DataInputStream data, int NumBytes)
	{
		int ReadInNum=0;
		try
		{
			for(int i=0;i<NumBytes;i++)
				ReadInNum+=data.readUnsignedByte()<<(8*i);
		}
		catch(IOException e1)
		{
			offscreen.drawString("IO Exception: "+e1.toString(), 10, 20);
		}
		return ReadInNum;
	}
	
	//Rendering functions
	public void update(Graphics screen) { paint(screen); }
	public void paint(Graphics g)
	{
		if(FaceLen==0) //An error on reading occured, so do nothing
		{
			g.drawImage(workspace, 0, 0, this); //Copy back buffer to front buffer
			return;
		}
		
		offscreen.clearRect(0,0,MovieWidth, MovieHeight); //Clear frame
		RotateWorld(.01f); //Rotate the model
		int xpoints[]=new int[3], ypoints[]=new int[3], S=0; //x&y arrays used to hold 2D projected vertex coordinates, S keeps track of the current vertex we are on (faster than array multiply lookup)
		float A[][]=matFinal, B[]=VertexArray; //Hold these locally for quicker access
		float CameraNormRotate[]=MatrixVectorMultiply(matWorld, CameraNorm); //Multiply world matrix against camera vector so we can more quickly determine which polygons are facing backwards
		for(int i=0;i<FaceLen;i++) //Loop through all polys
		{
			//If face is backwards, do not draw (back face culling) - we could store this value per face... but this would be against the point of the time tests
			if(BackFaceCulling && (Vec3Dot(Vec3Cross(Vec3SubFromArrNum(FaceArray[S],FaceArray[S+2]),Vec3SubFromArrNum(FaceArray[S+1],FaceArray[S+2])), CameraNormRotate))<0)
			{
				S+=3;
				continue;
			}
			
			//Calculate the 2D coordinates of our verticies
			int SA=FaceArray[S++]*3;
			float Q	= (A[0][3]*B[SA]+A[1][3]*B[SA+1]+A[2][3]*B[SA+2]+A[3][3]*1)*2;
			xpoints[0]	= (int)(((A[0][0]*B[SA]+A[1][0]*B[SA+1]+A[2][0]*B[SA+2]+A[3][0]*1)/Q+.5)*MovieWidth);
			ypoints[0]	= (int)(((A[0][1]*B[SA]+A[1][1]*B[SA+1]+A[2][1]*B[SA+2]+A[3][1]*1)/-Q+.5)*MovieHeight);
			SA=FaceArray[S++]*3;
			Q	= (A[0][3]*B[SA]+A[1][3]*B[SA+1]+A[2][3]*B[SA+2]+A[3][3]*1)*2;
			xpoints[1]	= (int)(((A[0][0]*B[SA]+A[1][0]*B[SA+1]+A[2][0]*B[SA+2]+A[3][0]*1)/Q+.5)*MovieWidth);
			ypoints[1]	= (int)(((A[0][1]*B[SA]+A[1][1]*B[SA+1]+A[2][1]*B[SA+2]+A[3][1]*1)/-Q+.5)*MovieHeight);
			SA=FaceArray[S++]*3;
			Q	= (A[0][3]*B[SA]+A[1][3]*B[SA+1]+A[2][3]*B[SA+2]+A[3][3]*1)*2;
			xpoints[2]	= (int)(((A[0][0]*B[SA]+A[1][0]*B[SA+1]+A[2][0]*B[SA+2]+A[3][0]*1)/Q+.5)*MovieWidth);
			ypoints[2]	= (int)(((A[0][1]*B[SA]+A[1][1]*B[SA+1]+A[2][1]*B[SA+2]+A[3][1]*1)/-Q+.5)*MovieHeight);

			//Draw polygons
			if(Polys)
			{
				offscreen.setColor(PolyColor);
				offscreen.fillPolygon(xpoints, ypoints, 3);				
			}
			if(WireFrame)
			{
				offscreen.setColor(WireColor);
				offscreen.drawPolygon(xpoints, ypoints, 3);				
			}
		}

		offscreen.drawString("w:"+WireFrame+", p:"+Polys+", b:"+BackFaceCulling+", #: "+FaceLen+", fps:"+Float.toString(GetFPS()), 10, 20); //Display the current FPS and toggle options
		g.drawImage(workspace, 0, 0, this); //Copy back buffer to front buffer
	}
	
	//Calculates FPS using an average over the last 2 seconds - See include/StdFps.cpp
	long FrameOn=0, TotalFrames=0, LastTotalFrames=0, OldTicks=0, StoreTicks=0;
	public float GetFPS()
	{
		FrameOn++;
		TotalFrames++;
		
		long CurrentTicks=System.currentTimeMillis();
		float AverageFPS=(1000.0f*FrameOn/(float)(CurrentTicks-OldTicks));

		if((CurrentTicks-OldTicks)>2000)
		{
			FrameOn-=LastTotalFrames;
			LastTotalFrames=FrameOn;
			OldTicks=StoreTicks;
			StoreTicks=CurrentTicks;
		}
		return AverageFPS;
	}
	
	//3D Math Used for cross product calculation
	float[] Vec3SubFromArrNum(int StartNum1, int StartNum2)
	{
		float[] F1=new float[3];
		float[] F2=new float[3];
		for(int i=0;i<3;i++)
		{
			F1[i]=VertexArray[StartNum1*3+i];
			F2[i]=VertexArray[StartNum2*3+i];
		}
		return Vec3Sub(F1, F2);
	}
	float Vec3Dot(float[] TheVec1, float[] TheVec2) { return TheVec1[0]*TheVec2[0]+TheVec1[1]*TheVec2[1]+TheVec1[2]*TheVec2[2]; }
	float[] Vec3Sub(float[] TheVec1, float[] TheVec2) { return new float[] {TheVec1[0]-TheVec2[0],TheVec1[1]-TheVec2[1],TheVec1[2]-TheVec2[2]}; }
	float[] Vec3Cross(float[] TheVec1, float[] TheVec2) { return new float[] {TheVec1[1]*TheVec2[2]-TheVec1[2]*TheVec2[1], TheVec1[2]*TheVec2[0]-TheVec1[0]*TheVec2[2],TheVec1[0]*TheVec2[1]-TheVec1[1]*TheVec2[0]}; }
	float[] MatrixVectorMultiply(float[][] A, float[] B)
	{
		float[] ReturnVec=new float[4];
		for(int i=0;i<4;i++)
			ReturnVec[i] = A[0][i]*B[0]+A[1][i]*B[1]+A[2][i]*B[2]+A[3][i]*1;
		return ReturnVec;
	}

	//Other 3D Math
	public void RotateWorld(float x)
	{
		CurRotate+=x;
		matWorld=new float[][] {{(float)Math.cos(CurRotate),0,(float)-Math.sin(CurRotate),0}, {0,1,0,0}, {(float)Math.sin(CurRotate),0,(float)Math.cos(CurRotate),0}, {0,0,0,1}};
		matFinal=MatrixMatrixMultiply(matWorld, matViewProj);
	}
	public float[][] MatrixMatrixMultiply(float A[][],float B[][])
	{
		float C[][]=new float[4][4];
		C[0][0] = A[0][0]*B[0][0]+A[0][1]*B[1][0]+A[0][2]*B[2][0]+A[0][3]*B[3][0];
		C[0][1] = A[0][0]*B[0][1]+A[0][1]*B[1][1]+A[0][2]*B[2][1]+A[0][3]*B[3][1];
		C[0][2] = A[0][0]*B[0][2]+A[0][1]*B[1][2]+A[0][2]*B[2][2]+A[0][3]*B[3][2];
		C[0][3] = A[0][0]*B[0][3]+A[0][1]*B[1][3]+A[0][2]*B[2][3]+A[0][3]*B[3][3];
		C[1][0] = A[1][0]*B[0][0]+A[1][1]*B[1][0]+A[1][2]*B[2][0]+A[1][3]*B[3][0];
		C[1][1] = A[1][0]*B[0][1]+A[1][1]*B[1][1]+A[1][2]*B[2][1]+A[1][3]*B[3][1];
		C[1][2] = A[1][0]*B[0][2]+A[1][1]*B[1][2]+A[1][2]*B[2][2]+A[1][3]*B[3][2];
		C[1][3] = A[1][0]*B[0][3]+A[1][1]*B[1][3]+A[1][2]*B[2][3]+A[1][3]*B[3][3];
		C[2][0] = A[2][0]*B[0][0]+A[2][1]*B[1][0]+A[2][2]*B[2][0]+A[2][3]*B[3][0];
		C[2][1] = A[2][0]*B[0][1]+A[2][1]*B[1][1]+A[2][2]*B[2][1]+A[2][3]*B[3][1];
		C[2][2] = A[2][0]*B[0][2]+A[2][1]*B[1][2]+A[2][2]*B[2][2]+A[2][3]*B[3][2];
		C[2][3] = A[2][0]*B[0][3]+A[2][1]*B[1][3]+A[2][2]*B[2][3]+A[2][3]*B[3][3];
		C[3][0] = A[3][0]*B[0][0]+A[3][1]*B[1][0]+A[3][2]*B[2][0]+A[3][3]*B[3][0];
		C[3][1] = A[3][0]*B[0][1]+A[3][1]*B[1][1]+A[3][2]*B[2][1]+A[3][3]*B[3][1];
		C[3][2] = A[3][0]*B[0][2]+A[3][1]*B[1][2]+A[3][2]*B[2][2]+A[3][3]*B[3][2];
		C[3][3] = A[3][0]*B[0][3]+A[3][1]*B[1][3]+A[3][2]*B[2][3]+A[3][3]*B[3][3];
		return C;
	}

	//Run the engine thread
	private Thread runner;
	public void start() { if(runner==null) { runner=new Thread(this); runner.start(); } }
	public void stop() { if(runner!=null) { runner.stop(); runner=null; } }
	public void run()
	{
		if(FaceLen!=0) //Only if file properly read, start main thread
			while(true)
				repaint();
	}
	
	//Handle key strokes
	public void keyPressed(KeyEvent e) { }
	public void keyReleased(KeyEvent e){}
	public void keyTyped(KeyEvent e)
	{
		switch(e.getKeyChar()) //Toggle options
		{
			case 'w':WireFrame=!WireFrame; break; //Toggle wireframe mode
			case 'p':Polys=!Polys;break; //Toggle poly mode
			case 'b':BackFaceCulling=!BackFaceCulling;break; //Toggle poly mode
		}
	}
}